# *****************************************************************
#
# Licensed Materials - Property of IBM
#
# (C) Copyright IBM Corp. 2019. All Rights Reserved.
#
# US Government Users Restricted Rights - Use, duplication or
# disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# *****************************************************************

from __future__ import print_function
import os
import time
import argparse
from scipy.sparse import load_npz
import numpy as np

# pai4sk explicit imports for snap_ml APIs
from pai4sk import MPI
from pai4sk import sml_io
from pai4sk import LogisticRegression as snapmlLR
from pai4sk.sml_metrics import log_loss as snapLOGLOSS
from pai4sk.sml_metrics import accuracy_score as snapACC

# pai4sk implicit API imports which use snap_ml or sklearn
from pai4sk.linear_model import LogisticRegression
from pai4sk.metrics import accuracy_score,log_loss

# pai4sk explicit imports for sklearn APIs
from pai4sk.linear_model import LogisticRegressionSklearn as sklearnLR
from pai4sk.metrics import accuracy_scoreSklearn,log_lossSklearn


defaultPath = os.getcwd()

(comm_size, rank) = MPI.Comm_get_info()

CLI=argparse.ArgumentParser()
CLI.add_argument(
   "--data_path",
   type=str,
   default=defaultPath
)
CLI.add_argument(
   "--device_ids",
   nargs="*",
   type=int,
   default=[]
)
CLI.add_argument(
    "--use_gpu",
    action = 'store_true',
    default=False
)
CLI.add_argument(
    "--skip_sklearn",
    action = 'store_true',
    default=False
)
CLI.add_argument(
   "--use_balanced_class_weights",
   action = 'store_true',
   default=False
)
args = CLI.parse_args()

if (rank == 0):
    print("[pai4sk] Data loading...")
t0 = time.time()
# Datasets paths (pai4sk distributed data loading)
train_filename_snap = args.data_path + "/data/criteo.kaggle2014.train.snap"
test_filename_snap = args.data_path + "/data/criteo.kaggle2014.test.snap"

## snapML
X_train,y_train = sml_io.load_snap_file(train_filename_snap)
X_test,y_test = sml_io.load_snap_file(test_filename_snap)

load_time_snap = time.time() - t0

# Create snapML Logistic Regression
if args.use_gpu:
  num_threads = 256
  cpu_gpu = "GPU"
else:
  num_threads = 8
  cpu_gpu = "CPU"


if args.use_balanced_class_weights:
  class_weight = "balanced"
else:
  class_weight = None

snapml_lr = snapmlLR(dual=True, use_gpu=args.use_gpu, device_ids=args.device_ids,
                     num_threads=num_threads, class_weight=class_weight)

if (rank == 0):
    print("[pai4sk] Training model on %s..." % (cpu_gpu))
# Fit the model and time it
snapml_t0 = time.time()
snapml_lr.fit(X_train,y_train)
snapml_time = time.time() - snapml_t0

if (rank == 0):
    print("[pai4sk] Running prediction...")
# Perform inference on test data
pred = snapml_lr.predict(X_test)
proba = snapml_lr.predict_proba(X_test)


# Compute accuracy
snapml_accuracy = snapACC(y_test, pred)
snapml_logloss = snapLOGLOSS(y_test, proba)


if (rank == 0):
    print("[pai4sk explicit] MPI process %d : Loading Time (s):  %.2f" % (rank, load_time_snap))
    print("[pai4sk explicit] MPI process %d : Accuracy:       %.4f" % (rank, snapml_accuracy))
    print("[pai4sk explicit] MPI process %d : Log loss:       %.4f" % (rank, snapml_logloss))
    print("[pai4sk explicit] MPI process %d : Training Time:       %.4f" % (rank, snapml_time))


lr = LogisticRegression(dual=True, use_gpu=args.use_gpu, device_ids=args.device_ids,
                     num_threads=num_threads, class_weight=class_weight)

t0 = time.time()
lr.fit(X_train,y_train)
lr_time = time.time() - t0

if (rank == 0):
    print("[pai4sk] Running prediction...")
# Perform inference on test data
pred = lr.predict(X_test)
proba = lr.predict_proba(X_test)


# Compute accuracy
accuracy = accuracy_score(y_test, pred)
logloss = log_loss(y_test, proba)

# Print off SnapML result
if (rank == 0):
    print("[pai4sk implicit] MPI process %d : Loading Time (s):  %.2f" % (rank, load_time_snap))
    print("[pai4sk implicit] MPI process %d : Accuracy: %.4f" % (rank, accuracy))
    print("[pai4sk implicit] MPI process %d : LogLoss = %.4f" % (rank, logloss))
    print("[pai4sk implicit] MPI process %d : Training time = %.4f" % (rank, lr_time))

## SKLEARN
if args.skip_sklearn:
    if (rank == 0):
        print('Execution using scikit-learn is skipped.')
    exit()

if (rank == 0):
    print("[sklearn] Data loading...")
t0 = time.time()
# Datasets paths (sklearn loading)
x_train_filename_sklearn = args.data_path + "/data/criteo.kaggle2014.X_train.npz"
y_train_filename_sklearn = args.data_path + "/data/criteo.kaggle2014.y_train.npy"
x_test_filename_sklearn = args.data_path + "/data/criteo.kaggle2014.X_test.npz"
y_test_filename_sklearn = args.data_path + "/data/criteo.kaggle2014.y_test.npy"

X_train = load_npz(x_train_filename_sklearn)
X_test  = load_npz(x_test_filename_sklearn)
y_train = np.load(y_train_filename_sklearn)
y_test  = np.load(y_test_filename_sklearn)

load_time_sklearn = time.time() - t0

# Import model
sklearn_lr = sklearnLR(dual=True, tol=0.001)

# Fit the model and time it
sklearn_t0 = time.time()
sklearn_lr_model = sklearn_lr.fit(X_train, y_train)
sklearn_time = time.time() - sklearn_t0

# Perform inference on test data
pred_proba = sklearn_lr.predict_proba(X_test)
pred = sklearn_lr.predict(X_test)

sklearn_accuracy = accuracy_scoreSklearn(y_test, pred) 
sklearn_logloss = log_lossSklearn(y_test, pred_proba)

if (rank == 0):
    print("[sklearn] [Data loading] Time = %.4f" % (load_time_sklearn))
    print("[sklearn] Accuracy = %.4f" % (sklearn_accuracy))
    print("[sklearn] LogLoss = %.4f" % (sklearn_logloss))
    print("[sklearn] Training time = %.4f" % (sklearn_time))


